// msg.c: Message display system for Bapple2.  Created on May 31st, 1997.

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dos.h>
#include <stdarg.h>

#include <allegro.h>
#include "bapple2.h"
#include "bap2dat.h"
#include "globals.h"
#include "msg.h"
#include "units.h"
#include "comm.h"

//char msg_color[MSG_TYPES] = { 1, 177, 112, 210, 177 };

RGB msg_color[MSG_TYPES] = {
	222, 222, 222,
	230, 206, 182,
	198, 206, 255,
	206, 190, 165,
	230, 206, 182
};

msg_buffer_t msg_buffer;

char cursoron = 0;
unsigned long blinktime = 0;

char last_message[4096];
int last_msg_type = 0;

// Message functions
void init_msg_system(void);
void add_msg(char type, char *text, ...);
void add_string(char type, char *text);
void process_messages(void);
void draw_messages(void);

// Command functions
void parse_command(char *text);
void invert_terrain(void);

// Misc utility functions
void remap_colors(BITMAP *bmp, PALETTE from, PALETTE to);

// From bapple2.cc:
extern void find_closest_free_tile(int *dx, int *dy);
extern void screen_shot(void);
extern void convert_to_linear(DATAFILE *datafile);
extern void render_world(void);

// From orders.cc:
extern void issue_order(unit_t *unit, char orders);
extern void free_path_nodes(unit_t *unit);
extern void destroy_all_units(void);

// From map.cc:
extern int save_map(char *filename);
extern int load_map(char *filename);
extern int load_vista(char *filename, int sealevel, int smoothing);
extern void fill_map_with_clipboard(void);
extern void random_paste(int num);
extern void destroy_windows(void);
extern int tileperiodicity;
extern int tileareaw, tileareah;

// WSOCK stuff
//extern Socket *s;
extern char Connect[80];
extern int Send(void *Buffer, int BufferLength, char *Address);

void init_msg_system(void)
{
	char buf[400];
	int i, j;
	FILE *fp;
#ifndef WIN32
	struct date d;
	struct time t;
#endif

	msg_buffer.initialized = 1;
	msg_buffer.log = 1;
    msg_buffer.num_msgs = 0;
	msg_buffer.top_msg = 0;

	if (msg_buffer.log)
	{
		fp = fopen("msg.log", "a");
#ifndef WIN32
		getdate(&d);
		gettime(&t);
		fprintf(fp, "\n---Message log opened on %d/%d/%d at %d:%02d%cM---\n\n", d.da_mon, d.da_day, d.da_year, (t.ti_hour > 12) ? t.ti_hour - 12 : t.ti_hour, t.ti_min, (t.ti_hour > 12) ? 'P' : 'A');
#else
		fprintf(fp, "\n---Message log opened---\n\n");
#endif
		fclose(fp);
	}

	// Set the last (entered) message to null
	last_message[0] = '\0';
	last_msg_type = MSG_COMPUTER;

	// Find the max message length, depending on both the font and the screen width
	for (i = 0; i < (400 - 2); i++)
		buf[i] = '\0';
	for (i = 0; i < (400 - 3); i++)
	{
		buf[i] = 'a';
		if (text_length(font, buf) > PhysicalWidth - 32 - text_length(font, "FPS: 00"))
			break;
	}
	msg_buffer.max_msg_width = i - 1;
//	msg_buffer.max_msg_width = 80;	// FIXME
	talk_buf[0] = '\0';
}

// Add a formatted message to the message buffer (using variable # of arguments)
void add_msg(char type, char *text, ...)
{
	va_list ap;
	char s[4096];
	int i, j;

	// Do the formatting and come up with one string
	va_start(ap, text);
	vsprintf(s, text, ap);
	va_end(ap);

	// If the message buffer is not yet initialized, use printf() instead
	if (msg_buffer.initialized != 1)
	{
		printf("%s\n", s);
		return;
	}

	// If the user wants to do the same message/command again, do it..
	if (((!strcmp(s, "> /!!")) || (!strcmp(s, "> /")) && (last_message[0])))
		add_string(last_msg_type, last_message);
	else		// Otherwise this COULD be a new message/command to be saved
	{
		if (type == MSG_NORMAL)			// Only save messages from the player!
		{
			strcpy(last_message, s);
			last_msg_type = type;
		}
		add_string(type, s);
	}
}

// Are we currently recursing into add_msg?  If so, don't log the message again
char recursing = 0;

// Add a single message string to the message buffer
void add_string(char type, char *text)
{
	msg_t *p;
	msg_t *backp;
	msg_t *msg;
	FILE *fp;
	char *moretxt = 0;
	char *savedtxt = 0;
	char add_at_end;
	int i, j, k;

	// If it's a command, parse that baby!
	if ((text[0] == '>') && (text[2] == '/'))
	{
		parse_command(&text[3]);
		// Don't redisplay player commands over and over on repeat requests (like /!!)
		if (!strcmp(last_message, text))
			return;
	}

	// Save the text, cause it gets modified
	savedtxt = (char *) malloc(strlen(text) + 1);
	strcpy(savedtxt, text);

	// Log the message, if logging is enabled
	if ((msg_buffer.log) && (!recursing))
	{
		fp = fopen("msg.log", "a");
//		if (msg_buffer.num_msgs >= MAX_MSGS)
//			fprintf(fp, "*Lost* ");
		fprintf(fp, "%s\n", text);
		fclose(fp);
	}

	// If the message buffer is full...
	if (msg_buffer.num_msgs >= MAX_MSGS)
	{
		// Get rid of the oldest (top) message to make room for the new one
		msg_buffer.top_msg->dietime = ms_time - 1;
		process_messages();
//		free(savedtxt);
//		return;
	}

	// Break down large messages into small ones, so they'll fit
	// on the screen ok
	while (text_length(font, text) > (SCREEN_W - (SCREEN_W / 5)))
	{
		moretxt = (char *) malloc(strlen(text) + 1);
		strcpy(moretxt, text);
		i = strlen(moretxt) - 1;
		// Keep searching backwards until the string is short enough
		while ((i > 0) && (text_length(font, moretxt) > (SCREEN_W - (SCREEN_W / 5))))
			moretxt[i--] = '\0';
		// Keep searching backwards until a space is found (word wrap, sorta)
		while ((i > 0) && moretxt[i] != ' ')
			i--;
		// If we searched all the way back to the beginning, screw it..
		// Just split the message in half
		if (i == 0)
			i = strlen(text) / 2;
		// Copy the correct 1st string to the moretxt buffer
		for (j = 0; j < i; j++)
			moretxt[j] = text[j];
		moretxt[++j] = '\0';
		// Yes, we are recursing..  set a flag to disable logging
		recursing = 1;
		// Add the new text to the message list
		add_msg(type, moretxt);
		// Allocate memory for the new text (only if necessary)
//		if (!finaltxt)
//	        finaltxt = (char *) malloc(strlen(text) + 1);
		// Create the new text from the old text
		for (k = 0; j < strlen(text); j++)
			text[k++] = text[j];
		text[k] = '\0';
		// Free the moretxt buffer
		if (moretxt)
			free(moretxt);
	}

	// Find the end of the list
	p = msg_buffer.top_msg;
	while (p)
	{
		backp = p;
		p = p->next;
	}

	// Set up the new message structure
	msg = (msg_t *) malloc(sizeof(msg_t));
	msg->type = type;
	msg->dietime = ms_time + 12000;
	msg->text = (char *) malloc(strlen(text) + 1);
	strcpy(msg->text, text);
	msg->next = 0;

	add_at_end = 1;
	// Add it to the list
	if (type == MSG_EXPENDABLE)
	{
        // Seek out and replace an existing message
		backp = msg_buffer.top_msg;
		for (p = msg_buffer.top_msg; p; p = p->next)
		{
			if (p->type == MSG_EXPENDABLE)
			{
				add_at_end = 0;
				msg->next = p->next;
				free(p->text);
                if (msg_buffer.top_msg == p)
				{
					free(p);
					msg_buffer.top_msg = msg;
				}
				else
				{
					free(p);
					backp->next = msg;
				}
				add_at_end = 0;
				break;
			}
			backp = p;
		}
	}

	if (add_at_end)
	{
		// Add it to the end of the list
		if (msg_buffer.top_msg)
			backp->next = msg;
		else
			msg_buffer.top_msg = msg;
		msg_buffer.num_msgs++;
	}

	// Free the memory allocated for the finaltxt buffer
//	free(finaltxt);

	// Restore the saved text
	strcpy(text, savedtxt);
	free(savedtxt);

	// We're not recursing anymore
	recursing = 0;
}

// Kill any messages that have expired
void process_messages(void)
{
	msg_t *p;
	msg_t *backp;

	p = msg_buffer.top_msg;
	backp = p;
	while (p)
	{
		if (ms_time > p->dietime)
		{
			if (p == msg_buffer.top_msg)
			{
				free(p->text);
				if (!p->next)
				{
					free(p);
					msg_buffer.top_msg = 0;
					msg_buffer.num_msgs--;
					break;
				}
				else
					msg_buffer.top_msg = p->next;
                free(p);
				p = msg_buffer.top_msg;
				msg_buffer.num_msgs--;
			}
			else
			{
				backp->next = p->next;
				free(p->text);
				free(p);
				msg_buffer.num_msgs--;
				p = backp;
			}
		}
		backp = p;
		p = p->next;
	}
}

void draw_messages(void)
{
	char buf[80];
	msg_t *p;
	int y, r, g, b;
	static char wastalking = 0;

	y = 32;
	p = msg_buffer.top_msg;

	// Draw the text
	while (p)
	{
		r = msg_color[p->type].r;
		g = msg_color[p->type].g;
		b = msg_color[p->type].b;
        textout(HiddenPage, font, p->text, 32, y, makecol(r, g, b));
		y += text_height(font) + 4;
		p = p->next;
	}

	// If the person is talking (typing), display a cursor
	if (talking)
	{
		if (!wastalking)
		{
			cursoron = 1;
			blinktime = ms_time + 200;
		}
		else
		if ((cursoron) && (ms_time > blinktime))
		{
            cursoron = 0;
			blinktime = ms_time + 200;
		}
		else
		if ((!cursoron) && (ms_time > blinktime))
		{
			cursoron = 1;
            blinktime = ms_time + 200;
		}

		if (cursoron)
		{
		talk_buf[cursor] = '_';
		talk_buf[cursor + 1] = '\0';
		}
		else
			talk_buf[cursor] = '\0';
		r = msg_color[MSG_NORMAL].r;
		g = msg_color[MSG_NORMAL].g;
		b = msg_color[MSG_NORMAL].b;
		textout(HiddenPage, font, talk_buf, 32, y, makecol(r, g, b));
	}
	wastalking = talking;
}

/* invert_terrain(): ------------------------------------------------------
 *   Inverts the terrain- works for land and water.  This will also
     [try to] relocate all of the units so that they remain on land.
 */
void invert_terrain(void)
{
    int x, y, i;

    for (y = 0; y < MAP_HEIGHT; y++)
        for (x = 0; x < MAP_WIDTH; x++)
        {
            // Change the terrain tiles
            switch (mpos(terrain, y, x))
            {
                case TERRAIN_WATER: mpos(terrain, y, x) = TERRAIN_LAND; break;
                case TERRAIN_LAND: mpos(terrain, y, x) = TERRAIN_WATER; break;
            }
            // Change the visible tiles
            switch (mpos(map, y, x))
            {
                case 26: if (rand() % 2) mpos(map, y, x) = 0; else mpos(map, y, x) = 1; break;
                case 0:
                case 1: mpos(map, y, x) = 26; break;

                case 20: mpos(map, y, x) = 2; break;
                case 2: mpos(map, y, x) = 20; break;

                case 21: mpos(map, y, x) = 11; break;
                case 27:
                case 28:
                case 11: mpos(map, y, x) = 21; break;

                case 16: mpos(map, y, x) = 15; break;
                case 15: mpos(map, y, x) = 16; break;

                case 14: mpos(map, y, x) = 17; break;
                case 17: mpos(map, y, x) = 14; break;

                case 12: mpos(map, y, x) = 18; break;
                case 18: mpos(map, y, x) = 12; break;

                case 19: mpos(map, y, x) = 13; break;
                case 13: mpos(map, y, x) = 19; break;

                case 25: mpos(map, y, x) = 24; break;
                case 24: mpos(map, y, x) = 25; break;
                case 22: mpos(map, y, x) = 23; break;
                case 23: mpos(map, y, x) = 22; break;
            }
        }
// Put the units where they can be happy
/*    for (i = 0; i < active_units; i++)
    {
        find_closest_free_tile(&unit[i].x, &unit[i].y);
        if (unit[i].path != NULL)
        {
            while (unit[i].path->Parent != NULL)
            {
                free(unit[i].path);
                unit[i].path = unit[i].path->Parent;
            }
            unit[i].path = unit[i].OPEN = unit[i].CLOSED = NULL;
        }
        issue_order(&unit[i], ORDER_GUARD);
    } */
}

void parse_command(char *text)
{
	int i, j, x, y;
	int argc;
	char argv[40][80];
	char buf[256];
	BITMAP *tmpbmp;
	PALETTE thepal, curpal, oldpal;
	struct MBASE_NODE *mnode;
	unit_t *u;
	thing_t *tmpthing;

	j = argc = 0;
	// Separate the command into separate words
	for (i = 0; i < strlen(text); i++)
	{
		if (text[i] == ' ')
		{
			argv[argc++][j] = '\0';
			j = 0;
		}
		else
			argv[argc][j++] = text[i];
	}
	argv[argc++][j] = '\0';

	// *** Process commands ***

	if (argc > 2) // Double-argument commands
	{
		if (!stricmp(argv[0], "loadvista") || !stricmp(argv[0], "lv"))
		{
			add_msg(MSG_COMPUTER, "All units have been destroyed.");
			destroy_all_units();
			// Load the map (use user-specified sea-level)
			if (argc > 3)
				load_vista(argv[1], atoi(argv[2]), atoi(argv[3]));
			else
				load_vista(argv[1], atoi(argv[2]), 0);	// no smoothing
		}
	}
	else
	if (argc > 1) // Single-argument commands
	{
		// Load a tile to replace the one under the player
		if (!stricmp(argv[0], "loadpic"))
		{
			tmpbmp = load_bitmap(argv[1], thepal);
			if (tmpbmp)
			{
				get_palette(curpal);
				remap_colors(tmpbmp, thepal, curpal);
				terraingfx[mpos(map, player.y, player.x)].dat = get_compiled_sprite(tmpbmp, is_planar_bitmap(screen));
				destroy_bitmap(tmpbmp);
			}
			else
				add_msg(MSG_COMPUTER, "Couldn't load that file.");
		}
		else
		if (!stricmp(argv[0], "loadpal"))
		{
			if (tmpbmp = load_bitmap(argv[1], thepal))
			{
                set_palette(thepal);
				create_rgb_table(&rgb_table, thepal, NULL);
				rgb_map = &rgb_table;
				destroy_bitmap(tmpbmp);
			}
			else
				add_msg(MSG_COMPUTER, "Couldn't load that file.");
		}
		else
		if (!stricmp(argv[0], "savemap"))
		{
			save_map(argv[1]);
		}
		else
		if (!stricmp(argv[0], "loadmap"))
		{
			add_msg(MSG_COMPUTER, "All units have been destroyed.");
			destroy_all_units();
			// Load the map
			load_map(argv[1]);
			// Search the datafile for a palette to load
			j = 0;
			for (i = 0; terraingfx[i].type != DAT_END; i++)
				if (terraingfx[i].type == DAT_PALETTE)
				{
					get_palette(oldpal);
	                set_palette((RGB *) terraingfx[i].dat);
					create_rgb_table(&rgb_table, (RGB *) terraingfx[i].dat, NULL);
					rgb_map = &rgb_table;
					get_palette(thepal);
//					TEXT_COLOR = makecol(223, 207, 239);
					j = 1;
					break;
				}
			if (!j)
			{
				add_msg(MSG_COMPUTER, "Error: Loaded tile datafile %s, but it didn't have a palette", tilefile);
				tilefile[0] = '\0';
				unload_datafile(terraingfx);
				terraingfx = 0;
			}
			else  // Search the datafile for bitmaps to have their palettes remapped
			{
				j = 0;
                for (i = 0; datafile[i].type != DAT_END; i++)
					if (datafile[i].type == DAT_BITMAP)
					{
						remap_colors((BITMAP *) datafile[i].dat, oldpal, thepal);
						j++;
					}

				tileperiodicity = 16;
				tileareaw = 16;
				tileareah = 28;

				add_msg(MSG_COMPUTER, "Loaded tile datafile %s, remapped %d bitmaps", tilefile, j);
			}
		}
		else
		if (!stricmp(argv[0], "loadvista") || !stricmp(argv[0], "lv"))
		{
			add_msg(MSG_COMPUTER, "All units have been destroyed.");
			destroy_all_units();
			// Load the map (automatically determine sea-level)
//			load_vista(argv[1], -1);
			load_vista(argv[1], 1, 0);
		}
		else
		if (!stricmp(argv[0], "loadgfx"))
		{
			load_terrain(argv[1]);
		}
		else
		if (!stricmp(argv[0], "attack"))
		{
			// The player has to be carrying a unit
			u = player.unit_carried;
			if (!u)
				return;
			// Turn the unit into a traitor if ordered to attack own base
			if (u->owner == atoi(argv[1]))
				u->owner++;
			// Send the unit to attack the specified enemy base
			for (mnode = mbase; mnode != NULL; mnode = mnode->next)
//				if ((mnode->health > 0) && (mnode->owner != myself->owner))
				if (mnode->owner == atoi(argv[1]))
				{
					desired_terrain = unit_terrain[u->type];
					x = mnode->x + 2;
					y = mnode->y + 2;
					// Find a position to take
					find_closest_free_tile(&x, &y);
		
					// Calculate a path to that position
		
					// First their tile must not be marked in unit_at though!
					tmpthing = mpos(unit_at, u->y, u->x);
					mpos(unit_at, u->y, u->x) = 0;
					// Calculate the path and store it with the unit
					u->path = (struct NODE *) FindPath((long)x,(long)y,(long)u->x,(long)u->y);
					// Now the unit tile can be remarked as being occupied
					mpos(unit_at, u->y, u->x) = tmpthing;
					if (u->path)
					{
						u->OPEN = OPEN;
						u->CLOSED = CLOSED;
						u->dx = x;
						u->dy = y;
						issue_order(u, ORDER_GOTO);
					}
					return;
				}
		}
		else
		if (!stricmp(argv[0], "toss") || !stricmp(argv[0], "random"))
		{
			random_paste(atoi(argv[1]));
		}
		else
		if (!stricmp(argv[0], "ping"))
		{
			if (comm_mode != COMM_NONE)
			{
				for (i = 0; i < active_enemies; i++)
					if (!stricmp(argv[1], enemy[i].name))
						j = i;
				buf[0] = PACKET_PING;
				Send(&buf[0], 1, enemy[j].addr);
				enemy[j].pingtime = ms_time;
				add_msg(MSG_COMPUTER, "Pinging %s...", enemy[j].name);
			}
		}
	}
	else	// No-argument commands
	{
		// Clone the player to a new enemy - FOR TESTING PURPOSES ONLY!!
		if (!stricmp(argv[0], "clone"))
		{
			add_msg(MSG_COMPUTER, "Created clone #%d", active_enemies);
			memcpy(&enemy[active_enemies], &player, sizeof(player_t));
			enemy[active_enemies].comm_state = CS_CONNECTED;
			active_enemies++;
		}
		else
		if (!stricmp(argv[0], "tt") || !stricmp(argv[0], "terraintype"))
		{
			display_terrain_type = !display_terrain_type;
			add_msg(MSG_COMPUTER, "Toggled display of terrain types.");
		}
		else
		if (!stricmp(argv[0], "grid"))
		{
			display_grid = !display_grid;
			add_msg(MSG_COMPUTER, "Toggled display of tile grid.");
		}
		else
		if (!stricmp(argv[0], "blur"))
		{
			motion_blur = !motion_blur;
			add_msg(MSG_COMPUTER, "Toggled motion blur effect.");
		}
		else
		if (!stricmp(argv[0], "showpath"))
		{
			showpath = !showpath;
			add_msg(MSG_COMPUTER, "Toggled display of path debug info.");
		}
		else
		if (!stricmp(argv[0], "fill"))
		{
			fill_map_with_clipboard();
			redraw_terrain_bmp = 1;
		}
		else
		if (!stricmp(argv[0], "render"))
		{
			render_world();
		}
		else
		if (!stricmp(argv[0], "invert"))
		{
			add_msg(MSG_COMPUTER, "Terrain inverted.");
			invert_terrain();
			redraw_terrain_bmp = 1;
			destroy_all_units();     // kill all units, otherwise game crashes :)
		}
		else
		if (!stricmp(argv[0], "la") || !stricmp(argv[0], "lastattacker"))
		{
			sprintf(buf, "Last attacker had ID #: %d", player.lastattacker);
			add_msg(MSG_COMPUTER, buf);
		}
		else
		if ((!stricmp(argv[0], "ss")) || (!stricmp(argv[0], "screenshot")))
		{
			add_msg(MSG_COMPUTER, "Screen shot taken.");
			screen_shot();
		}
		else
		if (!stricmp(argv[0], "baseregen"))
		{
			player.mbase->health = 10000;
		}
		else
		if (!stricmp(argv[0], "numunits"))
		{
			j = 0;
			for (u = units; u; u = u->next)
				j++;
			add_msg(MSG_COMPUTER, "Number of units in existence: %d", j);
		}
		else
		if (!stricmp(argv[0], "selfdestruct"))
		{
			player.health = 0;
		}
		else
		if ((!stricmp(argv[0], "killemall")) || (!stricmp(argv[0], "killunits")))
		{
			destroy_all_units();
		}
		else
		if (!stricmp(argv[0], "killwindows"))
		{
			destroy_windows();
		}
		else
		if (!stricmp(argv[0], "gfx_mode"))
		{
			switch (gfx_method)
			{
				case PAGE_FLIP: add_msg(MSG_COMPUTER, "%dx%d (virtual %dx%d) using page flipping", SCREEN_W, SCREEN_H, VIRTUAL_W, VIRTUAL_H); break;
				case DOUBLE_BUFFER: add_msg(MSG_COMPUTER, "%dx%d (virtual %dx%d) using double buffering", SCREEN_W, SCREEN_H, VIRTUAL_W, VIRTUAL_H); break;
				default: add_msg(MSG_COMPUTER, "Unknown rendering system #%d!  Tell Tom to fix this.", gfx_method);
			}
		}
		else
		if (!stricmp(argv[0], "ping"))
		{
			if (comm_mode != COMM_NONE)
			{
				buf[0] = PACKET_PING;
				Send(&buf[0], 1, enemy[0].addr);
				enemy[0].pingtime = ms_time;
				add_msg(MSG_COMPUTER, "Pinging %s...", enemy[0].name);
			}
		}
		else
		if (!stricmp(argv[0], "goto"))
		{
			if (player.unit_carried)
				player.unit_carried->orders = ORDER_GOTO;
		}
	}
}

void remap_colors(BITMAP *bmp, PALETTE from, PALETTE to)
{
	int i, x, y, p, best, bestindex, test;
	signed int colormap[256];

	for (i = 0; i < 256; i++)
		colormap[i] = -1;

	for (y = 0; y < bmp->h; y++)
		for (x = 0; x < bmp->w; x++)
		{
			p = getpixel(bmp, x, y);	// read a pixel from the bitmap
			if (colormap[p] == -1)		// If we have not found a match for this color..
			{
				best = 10000;
				bestindex = 0;
				for (i = 0; i < 256; i++)	// go through palettes looking for a close color
				{
					test = abs(from[p].r - to[i].r) + abs(from[p].g - to[i].g) + abs(from[p].b - to[i].b);
					if (test < best)
					{
						best = test;
						bestindex = i;
					}
				}
				colormap[p] = bestindex;
			}
			putpixel(bmp, x, y, colormap[p]);
		}
}
